unit uDrehen; { mk, 23.2.2004 }

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Math;
type
  TForm1 = class(TForm)
    bZeichnePolygon: TButton;
    lP0x: TLabel; lP0y: TLabel; lP1x: TLabel; lP1y: TLabel;
    lP2x: TLabel; lP2y: TLabel; lP3x: TLabel; lP3y: TLabel;
    eP0x: TEdit;  eP0y: TEdit;  eP1x: TEdit;  eP1y: TEdit;
    eP2x: TEdit;  eP2y: TEdit;  eP3x: TEdit;  eP3y: TEdit;
    lN: TLabel;  eN: TEdit;
    procedure bZeichnePolygonClick(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    P               : array of TPoint;  // Polygon
    linksgedrueckt  : boolean;          // linke Maustaste gedrckt
    rechtsgedrueckt : boolean;          // linke Maustaste gedrckt
    xa,ya           : integer;          // alte Mausposition
    alteFarbe       : TColor;
    Puffer          : TBitMap;          // Zeiger auf Puffer
  public
    { Public-Deklarationen }
  end;

var
  Form1: TForm1;

implementation
{$R *.dfm}

  function ccw(p0,p1,p2 : TPoint) : integer;
  var
    dx1,dx2,dy1,dy2 : integer;
  begin
    dx1 := p1.x - p0.x; dy1 := p1.y - p0.y;
    dx2 := p2.x - p0.x; dy2 := p2.y - p0.y;
    if dx1*dy2 > dy1*dx2 then result := +1;
    if dx1*dy2 < dy1*dx2 then result := -1;
    if dx1*dy2 = dy1*dx2 then
    begin
      if (dx1*dx2 < 0) or (dy1*dy2 < 0) then
        result := -1
      else
        if (dx1*dx1 + dy1*dy1) >= (dx2*dx2 + dy2*dy2) then
          result := 0
        else
          result := 1;
    end;
  end;

  function inside(Punkt : TPoint; P : array of TPoint) : boolean;
  var
    i,n   : integer;
    innen : boolean;
  begin
    n := Length(P);
    innen := true;
    i := 0;
    while (i < n-1) and innen do
    begin
      innen := ( ccw(P[i],P[i+1],Punkt) = +1);
      Inc(i);
    end;
    if innen then innen := ( ccw(P[n-1],P[0],Punkt) = +1);
    result := innen;
  end;

  function d(A,B : TPoint) : real;
  begin
    result := sqrt(sqr(B.X-A.X)+sqr(B.Y-A.Y));
  end;

  function winkel(P,S,W : TPoint) : real;
  var
    a,b,c,cosa : real;
  begin
    a := d(P,W); b := d(W,S); c := d(S,P);
    if b*c <> 0 then
    begin
      cosa := (b*b+c*c-a*a)/(2*b*c);
      if abs(cosa) < 1 then
        result := ccw(S,W,P)*arccos(cosa)
      else
        result := 0;
    end
    else
      result := 0;
  end;

  function Schwerpunkt(P : array of TPoint) : TPoint;
  var
    i,n   : integer;
    sx,sy : real;
  begin
    n := Length(P);
    sx := 0; sy := 0;
    for i := 0 to n-1 do
    begin
      sx := sx + P[i].X; sy := sy + P[i].Y;
    end;
    result.X := round(sx/n);
    result.Y := round(sy/n);
  end;

  function gedreht(P,D : TPoint; alpha : real) : TPoint;
  begin
    result.X := round(cos(alpha)*(P.X-D.X) - sin(alpha)*(P.Y-D.Y) + D.X);
    result.Y := round(sin(alpha)*(P.X-D.X) + cos(alpha)*(P.Y-D.Y) + D.Y);
  end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  linksgedrueckt := false; rechtsgedrueckt := false;
  Puffer := TBitMap.Create;
  Puffer.Width := width; Puffer.Height := height;
end;

procedure TForm1.bZeichnePolygonClick(Sender: TObject);
var
  i,n : integer;
  S   : TPoint;
begin
  // Ecken des Polygons eingeben
  n := StrToInt(eN.Text);
  setLength(P,n);
  if (n = 3) or (n=4) then
  begin
    P[0].X := StrToInt(eP0x.text); P[0].Y := StrToInt(eP0y.text);
    P[1].X := StrToInt(eP1x.text); P[1].Y := StrToInt(eP1y.text);
    P[2].X := StrToInt(eP2x.text); P[2].Y := StrToInt(eP2y.text);
  end;
  if n = 4 then
  begin
    P[3].X := StrToInt(eP3x.text); P[3].Y := StrToInt(eP3y.text);
  end;
  // Farbe des Polygons bestimmen
  Puffer.canvas.Brush.Color := canvas.Brush.Color;
  // Zeichenbereich lschen, dh. mit Brush-Farbe fllem
  Puffer.canvas.FillRect(Rect(0,0,width,height));
  alteFarbe := Puffer.canvas.Brush.Color;
  Puffer.canvas.Brush.Color := clRed;
  // Polygon zeichen
  Puffer.canvas.Polygon(P);
  // DEBUG Schwerpunkt zeichnen
  S := Schwerpunkt(P);
  Puffer.Canvas.Pixels[S.X,S.Y] := clWhite;
  // alte Farbe wieder einstellen
  Puffer.canvas.Brush.Color := alteFarbe;
  // Puffer zeichnen
  canvas.Draw(0,0,Puffer);
end;


procedure TForm1.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  Punkt : TPoint;
begin
  Punkt.X := x; Punkt.Y := y;
  if inside(Punkt,P) then
  begin
    if button = mbLeft  then linksgedrueckt  := true;
    if button = mbRight then rechtsgedrueckt := true;
    xa := x; ya := y;
  end;
end;

procedure TForm1.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  linksgedrueckt := false; rechtsgedrueckt := false;
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  i,dx,dy : integer;
  S,Px,W  : TPoint;
  alpha   : real;
begin
  if linksgedrueckt then
  begin
    // Differenzen bestimmen
    dx := x-xa; dy := y-ya;
    // Mauskoordinaten merken
    xa := x; ya := y;
    // neue Polygon-Koordinaten bestimmen
    for i := 0 to Length(P)-1 do
    begin
      P[i].X := P[i].X + dx;
      P[i].Y := P[i].Y + dy;
    end;
  end;
  if rechtsgedrueckt then
  begin
    S := Schwerpunkt(P); W.X := xa; W.Y := ya; Px.X := x; Px.Y := y;
    xa := x; ya := y;                  // Mauskoordinaten merken
    alpha := winkel(Px,S,W);
    // neue Polygon-Koordinaten bestimmen
    for i := 0 to Length(P)-1 do
    begin
      P[i] := gedreht(P[i],S,alpha);
    end;
  end;

  if linksgedrueckt or rechtsgedrueckt then
  begin
    // Zeichenbereich lschen, dh. mit Brush-Farbe fllem
    Puffer.canvas.FillRect(Rect(0,0,width,height));
    // Polygon zeichen
    alteFarbe := Puffer.canvas.Brush.Color;
    Puffer.canvas.Brush.Color := clRed;
    Puffer.canvas.Polygon(P);
    Puffer.canvas.Brush.Color := alteFarbe;
    // DEBUG Schwerpunkt zeichnen
    S := Schwerpunkt(P);
    Puffer.Canvas.Pixels[S.X,S.Y] := clWhite;
    // Puffer zeichnen
    canvas.Draw(0,0,Puffer);
  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  Puffer.Free;
end;

end.
